/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PHY_PHY_HH_
# define PHY_PHY_HH_

# include <wipal/tool/static_base.hh>
# include <wipal/tool/microseconds.hh>
# include <wipal/tool/endianness.hh>
# include <wipal/pcap/frame_descriptor.hh>

// Forward declarations.
namespace phy
{
  template <class B>	struct header;
  template <class B>	struct time;
  template <class B>	struct uint64_time;
}

// Virtual types.
namespace tool
{

  template <class B>
  struct types< phy::time<B> >
  {
    typedef undefined_type	impl_type;
  };

  template <class B>
  struct types< phy::uint64_time<B> >
  {
    typedef uint64_t		impl_type;
  };

  template <class B>
  struct types< phy::header<B> >
  {
    typedef undefined_type	time_type;
  };

} // End of namespace tool.


// Core declarations.
namespace phy
{

  enum
    {
      /**
       * Default precision for phy::eq_time_and_80211.
       *
       * 106 microseconds.  This is largest value one can afford when
       * synchronizing IEEE 802.11 traces.
       *
       * @see	phy::eq_time_and_80211,
       *	Yeo et al., A framework for wireless LAN monitoring and
       *	its applications.
       */
      eq_time_default_precision	= 106
    };

  /*----------.
  | phy::time |
  `----------*/

  /**
   * Base class for all PHY headers' embedded timing information.
   *
   * Some PHY headers (e.g. Prism, RadioTap, etc.) embed timing
   * information.  For instance Prism headers have @c mactime and @c
   * hosttime fields, RadioTap headers have a @c TSF field.  Those
   * fields must be updated during synchronization.  This abstract
   * class provides a common interface so WiPal's synchronization
   * functions work whatever the PHY header.  When implementing
   * a custom PHY header, subclass this to export timing information
   * to WiPal's internals.
   *
   * Some PHY headers hold precise timestamps indicating when frames
   * are recorded on the medium (e.g. @c mactime for Prism headers).
   * WiPal may use this instead of PCAP timestamps to enable a better
   * synchronization.  Such timestamps must be of type uint32_t,
   * uint64_t, or struct timeval.  When implementing a custom PHY
   * header, define @c impl_type to the type you want to use and make
   * timestamps available to WiPal with the @c get() method.  (Use
   * WiPal's virtual types mechanism to define @c impl_type).
   *
   * If your custom PHY header does not provide such precise
   * timestamps, define @c impl_type to whatever type you want
   * (e.g. an empty struct).  WiPal's synchronization function will
   * then use PCAP timestamps for synchronization.
   *
   * @see phy::empty_time, prism::time, phy::empty_header, prism::header
   */
  template <class Bottom = tool::bottom>
  struct time: WP_INHERIT(public tool::static_base, time<Bottom>)
  {
    /// @name Static hierarchies related types
    //@{
    typedef WP_GET_EXACT(Bottom, time<Bottom>)	exact_type;
    typedef WP_TYPE(impl_type, exact_type)	impl_type;
    //@}

    /**
     * Retrieve the timestamp to use for synchronization.
     *
     * This method calls @c get_impl() which must be defined in
     * subclasses.  Do not override it.  Define @c get_impl() in
     * subclasses instead.
     *
     * @see phy::time
     */
    impl_type	get() const;

    /**
     * Add a microsecond count to the current time value.
     *
     * This methods calls @c increment() which must be defined in
     * subclasses.  Do not override it.  Define @c increment() in
     * subclasses instead.
     *
     * @c increment() must return a reference to itself.
     */
    exact_type&	operator += (tool::microseconds microseconds);

    /// Add a microsecond count to the current time value.  @see operator+=()
    exact_type	operator +  (tool::microseconds microseconds) const;

  protected:
    ~time(); // phy::time is abstract!
  };

  /*-----------------.
  | phy::uint64_time |
  `-----------------*/

  /// Timing information for PHY headers that just include an 64 bit timer.
  template <class Bottom = tool::bottom>
  struct uint64_time: WP_INHERIT(public phy::time, uint64_time<Bottom>)
  {
    /// @name Static hierarchies related types and methods.
    //@{
    typedef WP_GET_EXACT(Bottom, uint64_time<Bottom>)	exact_type;
    typedef WP_TYPE(impl_type, exact_type)		impl_type;

    impl_type	get_impl() const;
    exact_type&	increment(const tool::microseconds& microseconds);
    //@}

    /// Default constructor.  Initialize the microseconds count to zero.
    uint64_time();

    /// Initialize the microseconds count according to the given argument.
    uint64_time(uint64_t ms_count);

  private:
    uint64_t	ms_count_;
  };

  /*------------.
  | phy::header |
  `------------*/

  /**
   * Base class for all PHY headers.
   *
   * All PHY headers shall subclass this class.  It provides a common
   * interface so WiPal's internals work whatever the PHY header.
   *
   * To subclass, you must define the @c time_type virtual type,
   * implement @c len_impl(), @c time_get_impl(), @c time_set_impl(),
   * and @c eq_time().
   *
   * @c time_type is the type of the PHY header's timing information.
   * Use phy::empty_time if your PHY header has no timing information.
   *
   * @c time_get_impl() and @c time_set_impl() retrieve and set this
   * PHY header's timing information.
   *
   * @c eq_time() indicates whether two frames are equals with respect
   * to timestamps.  If no timestamps are available, implement it by
   * comparing PCAP timestamps.
   *
   * @c len_impl() returns the size of the PHY header.
   *
   * @see prism::header, phy::empty_header
   */
  template <class Bottom = tool::bottom>
  struct header: WP_INHERIT(public tool::static_base, header<Bottom>)
  {
    /// @name Static hierarchies related types
    //@{
    typedef WP_GET_EXACT(Bottom, header<Bottom>)	exact_type;
    typedef WP_TYPE(time_type, exact_type)		time_type;
    //@}

    /**
     * Get the PHY header length.
     *
     * Never override this method.  Define @c len_impl() in
     * subclasses instead.
     */
    size_t	len(size_t caplen, bool swapped) const;

    /**
     * Get the embedded IEEE 802.11 frame.
     *
     * Return 0 if len() > caplen.
     *
     * This method relies on @c len().  You should not need to
     * override it if your subclasses define a proper @c len_impl().
     */
    const void*	decapsulate(size_t caplen, bool swapped) const;

    /**
     * Extract timing information from the PHY header.
     *
     * Never override this method.  Define @c time_get_impl() in
     * subclasses instead.
     */
    time_type	time_get(bool swapped) const;

    /**
     * Set the PHY header's timing information.
     *
     * Never override this method.  Define @c time_set_impl() in
     * subclasses instead.
     */
    void	time_set(const time_type& time, bool swapped);

    /**
     * Compare the IEEE 802.11 bytes beyond the PHY header.
     *
     * Return whether the two IEEE 802.11 frames embedded inside the
     * PHY headers are equals.  This function <b>does not compare</b>
     * the PHY headers.
     *
     * In case there is not enough bytes to decide whether frames are
     * equal (caplen < sizeof prism::header), consider they are not.
     * This is a conservative approach: when merging two traces, this
     * will lead to both frames having their entry in the merged trace.
     *
     * This method relies on @c decapsulate().  You should not need to
     * override it if your subclasses define a proper @c
     * decapsulate_impl().
     */
    template <class D1, class D2>
    static bool	eq_80211(const pcapxx::frame_descriptor<D1>&	lhs,
			 const pcapxx::frame_descriptor<D2>&	rhs,
			 tool::endian::endianness		phy_end =
				tool::endian::guess);


    /**
     * Compare the IEEE 802.11 bytes beyond PHY headers, but also
     * check timestamps.
     *
     * Return whether the two IEEE 802.11 frames embedded inside the
     * PHY headers are equals.  Also compare the timestamps of PHY
     * headers if available (or the PCAP timestamps else).  When the
     * time difference between each header's field is less than @c
     * precision, then consider the field equals.
     *
     * In case there is not enough bytes to decide whether frames are
     * equal (caplen < sizeof prism::header), consider they are not.
     * This is a conservative approach: when merging two traces, this
     * will lead to both frames having their entry in the merged trace.
     *
     * This method relies on @c eq_time().  You should not need to
     * override it if your subclasses define a proper @c eq_time().
     */
    template <class D1, class D2>
    static bool	eq_time_and_80211(const pcapxx::frame_descriptor<D1>&	lhs,
				  const pcapxx::frame_descriptor<D2>&	rhs,
				  tool::endian::endianness phy_end =
					tool::endian::guess,
				  unsigned precision =
					phy::eq_time_default_precision);

  protected:
    ~header(); // phy::header is abstract!
  };

} // End of namespace phy.

# include "phy.hxx"

#endif // ! PHY_PHY_HH_
