/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_STATS_MISSED_ESTIMATIONS_HXX_
# define WIFI_FRAME_STATS_MISSED_ESTIMATIONS_HXX_

# include <wipal/wifi/frame/stats/missed_estimations.hh>
# include <wipal/wifi/frame/frame.hh>
# include <wipal/wifi/frame/mgt.hh>

namespace wifi
{

  namespace frame
  {

    namespace stats
    {


      inline
      missed_estimations::missed_estimations(const simple_counters& c):
	missed_beacons_ (0),
	missed_ (0),
	c_ (&c)
      {
      }

      inline
      void
      missed_estimations::account_beacon(const tool::microseconds&	tstamp,
					 const addr&			src)
      {
	beacons_map::iterator	i = beacons_.find(src);

	if (i == beacons_.end())
	  {
	    beacons_.insert(std::make_pair(src, tstamp));
	    return;
	  }

	const tool::microseconds	dt = tstamp - i->second;

	if (dt < 0)
	  std::cerr << "WARNING: Time-travelling beacon!" << std::endl;
	else
	  {
	    const unsigned		delta = dt.get_div_by(beacon_interval);

	    /* Due to timers imprecision It is usual delta is zero.
	     * e.g.  if (tstamp - i->second) is 102396 or 101299 and
	     * beacon_interval is 102400(those are real empirical
	     * values).
	     */
	    if (1 < delta)
	      missed_beacons_ += delta - 1;
	  }

	i->second = tstamp;
      }

      inline
      void
      missed_estimations::account_miss()
      {
	++missed_;
      }

      inline
      void
      missed_estimations::account_gap(unsigned gap_length)
      {
	missed_ += gap_length;
      }

      inline
      std::ostream&
      missed_estimations::print(std::ostream& o) const
      {
	const unsigned	stc = c_->subtype_count(type::management,
						mgt::subtype::beacon);

	o << "estimated missed beacons: " << missed_beacons_ << '\n'
	  << "estimated missed beacons ratio: "
	  <<   (double (missed_beacons_) / (missed_beacons_ + stc)) << '\n'
	  << "estimated missed frames: " << missed_ << '\n'
	  << "estimated missed frames ratio: "
	  <<   (double (missed_) / (missed_ + c_->total())) << std::endl;
	return o;
      }

    } // End of namespace wifi::frame::stats.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_STATS_MISSED_ESTIMATIONS_HXX_
