/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_STATS_STATS_HH_
# define WIFI_FRAME_STATS_STATS_HH_

extern "C"
{
# include <stdint.h>
}

# include <map>
# include <utility>

# include <boost/tuple/tuple.hpp>
# include <boost/tuple/tuple_comparison.hpp>
# include <boost/optional.hpp>

# include <wipal/wifi/frame/stats/delimiter_frames.hh>
# include <wipal/wifi/frame/stats/simple_counters.hh>
# include <wipal/wifi/frame/stats/gap_lengths.hh>
# include <wipal/wifi/frame/stats/missed_estimations.hh>
# include <wipal/wifi/frame/stats/tfi_plot.hh>
# include <wipal/wifi/frame/stats/bss_stats.hh>
# include <wipal/wifi/frame/stats/ssid_stats.hh>
# include <wipal/wifi/frame/stats/activity.hh>
# include <wipal/wifi/frame/stats/growth.hh>
# include <wipal/tool/microseconds.hh>
# include <wipal/wifi/addr.hh>
# include <wipal/wifi/frame/seqctl.hh>
# include <wipal/wifi/frame/dissector/default_hooks.hh>
# include <wipal/wifi/frame/mgt.hh>

namespace wifi
{

  namespace frame
  {

    namespace stats
    {

      struct stats: public abstract_stats
      {
	stats();

	template <class PhyHeader, class Frame>
	void
	account(const Frame& f);

	void				restart();
	std::ostream&			print(std::ostream& o) const;

	enum { growth_epoch = 60 * 1000 * 1000 }; // 60 seconds.

	enum transmitter_type { unknown, client, ap };


	typedef growth<addr, growth_epoch>		addr_growth;
	typedef growth<addr, growth_epoch>		bssid_growth;
	typedef growth<addr, growth_epoch>		ibssid_growth;
	typedef growth<std::string, growth_epoch>	ssid_growth;
	typedef growth<addr, growth_epoch>		ap_growth;

	const delimiter_frames&		delimiter_frames_get() const;
	const simple_counters&		simple_counters_get() const;
	const gap_lengths&		gap_lengths_get() const;
	const missed_estimations&	missed_estimations_get() const;
	const tfi_plot&			tfi_plot_get() const;
	const bss_stats&		bss_stats_get() const;
	const ssid_stats&		ssid_stats_get() const;
	const activity&			activity_get() const;
	const addr_growth&		addr_growth_get() const;
	const bssid_growth&		bssid_growth_get() const;
	const ibssid_growth&		ibssid_growth_get() const;
	const ssid_growth&		ssid_growth_get() const;
	const ap_growth&		ap_growth_get() const;

      private:
	struct beacon_body
	{
	  uint64_t	timestamp;
	  uint16_t	beacon_interval;
	  uint16_t	capability;
	  uint8_t	ssid_id;
	  uint8_t	ssid_len;
	  uint8_t	ssid[];
	} __attribute__((packed));

	enum
	  {
	    state_lifetime = 60 * 1000 * 1000, // 60 second.
	    min_frame_sending_time = 106
	  };

	struct state
	{
	  state(unsigned			last_frame_id,
		const seqctl&			last_sc,
		const tool::microseconds&	last_rx_stamp);

	  unsigned		last_frame_id;
	  seqctl		last_sc;
	  tool::microseconds	last_rx_stamp;
	};

	typedef boost::tuple<addr, addr, uint8_t>	state_id;
	typedef std::map<state_id, state>		state_map;

	typedef std::pair<const mgt::header*, size_t>	beacon_header;

	struct hooks: public dissector_default_hooks
	{
	  hooks(stats& parent);

	  void	invalid_type_or_subtype_hook(const void*, size_t,
					     dissector_status::status);
	  void	addr_hook(const void*, size_t, unsigned, const addr&);
	  void	seq_ctl_hook(const void*, size_t, unsigned, unsigned);
	  void	qos_ctl_hook(const void*, size_t,
			     unsigned, bool, unsigned, unsigned);
	  void	beacon_hook(const mgt::header*, size_t);

	  typedef boost::optional<seqctl>		optional_seqctl;
	  typedef boost::optional<state_id>		optional_state_id;
	  typedef boost::optional<beacon_header>	optional_beacon_header;

	  unsigned			type;
	  unsigned			subtype;
	  bool				retx;
	  optional_seqctl		sc;
	  optional_state_id		id;
	  transmitter_type		txer_type;
	  optional_beacon_header	beacon;

	private:
	  const addr*			receiver;
	  stats*			parent;
	};

      private:
	typedef dissector<hooks>		stats_dissector;

	template <class Frame>
	void
	update_states(const Frame& f, const stats_dissector& d);

	template <class PhyHeader, class Frame>
	void
	account_frame(const Frame& f, const stats_dissector& d);


	void		account_beacon(const tool::microseconds&	tstamp,
				       const mgt::header*		header,
				       size_t				caplen,
				       bool				ibss);
	void		account_miss();
	void		account_gap(unsigned gaplen);
	void		account_expired();
	void		account_out_of_order();
	void		account_duplicate();
	void		account_seq_anomaly();
	void		account_addr(const addr& a);

	state_map		states_;
	std::set<addr>		aps_;

	delimiter_frames	df_;
	simple_counters		sc_;
	gap_lengths		gl_;
	missed_estimations	me_;
	tfi_plot		tp_;
	bss_stats		bs_;
	ssid_stats		ss_;
	activity		ay_;
	addr_growth		ag_;
	bssid_growth		bg_;
	bssid_growth		ig_;
	ssid_growth		sg_;
	ap_growth		pg_;
      };

    } // End of namespace wifi::frame::stats.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

# include "stats.hxx"

#endif // ! WIFI_FRAME_STATS_STATS_HH_
