/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_TRACES_SIMILARITY_HXX_
# define WIFI_TRACES_SIMILARITY_HXX_

# include "similarity.hh"

# include <boost/tuple/tuple_comparison.hpp>

# include <wipal/wifi/frame/filter/non_noisy_prism.hh>
# include <wipal/wifi/frame/filter/uniquely_identifiable.hh>

namespace wifi
{

  namespace traces
  {

    /// Internal helpers for similarity().
    namespace similarity_helpers
    {

      template <class I>
      void
      info(unsigned i, const I& t)
      {
	// Avoid unused argument warngings when not WP_ENABLE_INFO.
	(void) i;
	(void) t;

# ifdef WP_ENABLE_INFO
	std::cerr << "INFO: Trace " << i << " is " << t->file_name()
		  << std::endl;
# endif // WP_ENABLE_INFO
      }

      template <class Result, class Intersector, class I1, class I2>
      unsigned
      score(Result& r, Intersector& i, const I1& lhs, const I2& rhs)
      {
	unsigned	count = 0;

	typename Intersector::const_iterator it = i.begin();
	while (it != i.end() and
	       it.rhs_uframe_count() < similarity_score_limit)
	  {
	    ++it;
	    ++count;
	  }
	if (it != i.end() and count > 0)
	  --count;

	r.insert(boost::make_tuple(count, lhs, rhs));
# ifdef WP_ENABLE_INFO
	std::cerr << "INFO: Score for " << lhs->file_name()
		  << " and " << rhs->file_name()
		  << " is " << count << '.' << std::endl;
# endif // WP_ENABLE_INFO

	return count;
      }

      template <class U, class HT, template <class, class, class> class I,
		class C>
      std::set< boost::tuple<unsigned, typename C::iterator,
			               typename C::iterator> >
      ignore_prism(C&				traces,
		   addr_mapping&		mapping,
		   tool::endian::endianness	endianness)
      {
	using frame::filter::uniquely_identifiable;

	typedef typename C::iterator				trace_iterator;
	typedef typename C::value_type::const_iterator		frame_iterator;
	typedef uniquely_identifiable<U, frame_iterator, HT>	ui;
	typedef typename ui::const_iterator			ui_iterator;
	typedef I<ui_iterator, ui_iterator, tool::bottom>	intersector;
	typedef typename intersector::const_iterator		i_iterator;

	typedef std::set< boost::tuple<unsigned, typename C::iterator,
						 typename C::iterator> >
								result;

	result		r;

	if (traces.empty())
	  return r;

	trace_iterator	t = traces.begin();
	trace_iterator	lhs, rhs;

	while (lhs = t++, rhs = t, rhs != traces.end())
	  {
	    info(1, lhs);

	    ui			lhs_u (lhs->begin(), lhs->end(),
				       mapping, endianness);
	    intersector		i (lhs_u.begin(), lhs_u.end(),
				   lhs_u.begin(), lhs_u.end());
	    do
	      {
		info(2, rhs);

		ui	rhs_u (rhs->begin(), rhs->end(),
			       mapping, endianness);

		i.reset_rhs(rhs_u.begin(), rhs_u.end());
		score(r, i, lhs, rhs);
	      }
	    while (++rhs != traces.end());
	  }
	return r;
      }

      template <class U, class HT, template <class, class, class> class I,
		class C>
      std::set< boost::tuple<unsigned, typename C::iterator,
			               typename C::iterator> >
      filter_prism(C&				traces,
		   addr_mapping&		mapping,
		   tool::endian::endianness	endianness)
      {
	using frame::filter::non_noisy_prism;
	using frame::filter::uniquely_identifiable;

	typedef typename C::iterator				trace_iterator;
	typedef typename C::value_type::const_iterator		frame_iterator;
	typedef non_noisy_prism<frame_iterator>			nnp;
	typedef typename nnp::const_iterator			nnp_iterator;
	typedef uniquely_identifiable<U, nnp_iterator, HT>	ui;
	typedef typename ui::const_iterator			ui_iterator;
	typedef I<ui_iterator, ui_iterator, tool::bottom>	intersector;
	typedef typename intersector::const_iterator		i_iterator;

	typedef std::set< boost::tuple<unsigned, typename C::iterator,
						 typename C::iterator> >
								result;

	result		r;

	if (traces.empty())
	  return r;

	trace_iterator	t = traces.begin();
	trace_iterator	lhs, rhs;

	while (lhs = t++, rhs = t, rhs != traces.end())
	  {
	    info(1, lhs);

	    nnp			lhs_n (lhs->begin(), lhs->end());
	    ui			lhs_u (lhs_n.begin(), lhs_n.end(),
				       mapping, endianness);
	    intersector		i (lhs_u.begin(), lhs_u.end(),
				   lhs_u.begin(), lhs_u.end());
	    do
	      {
		info(2, rhs);

		nnp	rhs_n (rhs->begin(), rhs->end());
		ui	rhs_u (rhs_n.begin(), rhs_n.end(),
			       mapping, endianness);

		i.reset_rhs(rhs_u.begin(), rhs_u.end());
		score(r, i, lhs, rhs);
	      }
	    while (++rhs != traces.end());
	  }
	return r;
      }

    } // End of namespace wifi::frame::traces::similarity_helpers.


    template <class U, class P, template <class, class, class> class I,
	      class C>
    std::set< boost::tuple<unsigned, typename C::iterator,
				     typename C::iterator> >
    similarity(C&			traces,
	       addr_mapping&		mapping,
	       bool			filter_noisy_prism,
	       tool::endian::endianness	endianness)
    {
      return filter_noisy_prism?
	similarity_helpers::filter_prism<U, P, I>(traces, mapping, endianness):
	similarity_helpers::ignore_prism<U, P, I>(traces, mapping, endianness);
    }

  } // End of namespace wifi::traces.

} // End of namespace wifi.

#endif // ! WIFI_TRACES_SIMILARITY_HXX_
